/*
 *   Copyright 2011 Institute of Communication and Computer Systems (ICCS) 
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */

#include <stdio.h>
#include <dmmlib/dmmlib.h>
#include "other.h"
#ifdef HAVE_LOCKS
#include "posix_lock.h"
#endif /* HAVE_LOCKS */
#if defined (COALESCING_FIXED) || defined (COALESCING_VARIABLE)
#include "coalesce.h"
#endif /* COALESCING_FIXED || COALESCING_VARIABLE */
#include "linked_lists/add_block.h"
#ifdef WITH_ADAPTIVITY
#include "dmm_adaptor.h"
#endif /* WITH_ADAPTIVITY */
#include "block_header.h"

void custom_ahfree(allocator_t *allocator, heap_t* heap, void *ptr) {
    size_t size;
#ifdef WITH_FIXED_LISTS
    int fixed_list_id, i;
    maptable_node_t *current_maptable_node;
#endif /* WITH_FIXED_LISTS */
#if defined (COALESCING_FIXED) || defined (COALESCING_VARIABLE)
    bool coalesced;
#endif /* COALESCING_FIXED || COALESCING_VARIABLE */

    if(ptr == NULL) {
        return;
    }

    size = get_size(ptr);

#ifdef HAVE_LOCKS	
    posix_lock(heap);
#endif /* HAVE_LOCKS */

#ifdef WITH_STATS
    /* Update Stats */
#ifdef REQUEST_SIZE_INFO
    heap->dmm_stats.mem_requested -= get_requested_size(ptr);
#endif /* REQUEST_SIZE_INFO  */
    heap->dmm_stats.mem_used -= HEADER_SIZE + size;
    heap->dmm_stats.live_objects -= 1;
    heap->dmm_stats.num_free += 1;
    /* End of Stats */
#endif /* WITH_STATS */

#ifdef COUNT_HOPS
    /* Initialize the hops counter */
    heap->dmm_stats.total_hops = 0;
#endif /* COUNT_HOPS */

#ifdef FUTURE_FEATURES
#ifdef COUNT_HOPS
    remove_block(heap, &ptr, &heap->used_blocks_head);
#else
    remove_block(&ptr, &heap->used_blocks_head);
#endif /* COUNT_HOPS */
#endif /* FUTURE_FEATURES */

#if defined (COALESCING_FIXED) || defined (COALESCING_VARIABLE)
    coalesced = coalesce(allocator, heap, ptr);
#else
    mark_free(allocator, ptr);
#endif /* COALESCING_FIXED || COALESCING_VARIABLE */

#ifdef WITH_FIXED_LISTS
    /* Check if the block could be put in a fixed list */
    fixed_list_id = map_size_to_list(heap, size);

    if(fixed_list_id != -1) {
        current_maptable_node = heap->maptable_head;		
        if(fixed_list_id != 0) {
            for(i = 1; i < fixed_list_id; i++) {
                current_maptable_node = current_maptable_node->next;
            }
        }
#ifdef COUNT_HOPS
#ifdef FIFO_SORT_POLICY
        add_block(heap, &ptr, &current_maptable_node->fixed_list_head, &current_maptable_node->fixed_list_tail);
#else /* FIFO_SORT_POLICY */
        add_block(heap, &ptr, &current_maptable_node->fixed_list_head);
#endif /* FIFO_SORT_POLICY */
#else /* COUNT_HOPS */
#ifdef FIFO_SORT_POLICY
        add_block(&ptr, &current_maptable_node->fixed_list_head, &current_maptable_node->fixed_list_tail);
#else /* FIFO_SORT_POLICY */
        add_block(&ptr, &current_maptable_node->fixed_list_head);
#endif /* FIFO_SORT_POLICY */
#endif /* COUNT_HOPS */
    } else { /* put it in the free list */
#endif /* WITH_FIXED_LISTS */
#if defined (COALESCING_FIXED) || defined (COALESCING_VARIABLE)
        /* The block should be added to the free list only if it is not
         * coalesced
         */
        if(!coalesced) {
#ifdef COUNT_HOPS
#ifdef FIFO_SORT_POLICY
            add_block(heap, &ptr, &heap->free_list_head, &heap->free_list_tail);
#else /* FIFO_SORT_POLICY */
            add_block(heap, &ptr, &heap->free_list_head);
#endif /* FIFO_SORT_POLICY */
#else /* COUNT_HOPS */
#ifdef FIFO_SORT_POLICY
            add_block(&ptr, &heap->free_list_head, &heap->free_list_tail);
#else /* FIFO_SORT_POLICY */
            add_block(&ptr, &heap->free_list_head);
#endif /* FIFO_SORT_POLICY */
#endif /* COUNT_HOPS */
        }
#else /* COALESCING_FIXED || COALESCING_VARIABLE */
#ifdef COUNT_HOPS
#ifdef FIFO_SORT_POLICY
        add_block(heap, &ptr, &heap->free_list_head, &heap->free_list_tail);
#else /* FIFO_SORT_POLICY */
        add_block(heap, &ptr, &heap->free_list_head);
#endif /* FIFO_SORT_POLICY */
#else /* COUNT_HOPS */
#ifdef FIFO_SORT_POLICY
        add_block(&ptr, &heap->free_list_head, &heap->free_list_tail);
#else /* FIFO_SORT_POLICY */
        add_block(&ptr, &heap->free_list_head);
#endif /* FIFO_SORT_POLICY */
#endif /* COUNT_HOPS */
#endif /* COALESCING_FIXED || COALESCING_VARIABLE */
#ifdef WITH_FIXED_LISTS
    }
#endif /* WITH_FIXED_LISTS */

#ifdef WITH_ADAPTIVITY
    /* Refresh the state of the heap allocator if a certain number of
     * free's has been served already
     */
    /* TODO Define 100 as a constant */
    if(heap->dmm_stats.num_free % 100) {
        free_state_refresh(heap);
    }
#endif /* WITH_ADAPTIVITY */

#ifdef HAVE_LOCKS
    posix_unlock(heap);
#endif /* HAVE_LOCKS */
}
