/*
 *   Copyright 2011 Institute of Communication and Computer Systems (ICCS) 
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */

#include "block_header.h"
#include "other.h"
#include "linked_lists/add_block.h"
#include "split.h"

void split(allocator_t *allocator, heap_t *heap, void *ptr,
        size_t req_size) {
    size_t new_size;
    size_t min_split_size;
    void *new_block;
    size_t new_block_size;
#ifdef WITH_FIXED_LISTS
    int fixed_list_id, i;
    maptable_node_t *current_maptable_node;
#endif /* WITH_FIXED_LISTS */
#ifdef COALESCE_AFTER_SPLIT
    size_t max_coal_size;
    void *next_block;
    size_t current_next_size;

    next_block = get_dlnext(allocator, ptr);
#endif /* COALESCE_AFTER_SPLIT */

    /* Check what would be the size of the new block if we split the current
     * one.
     * Note: new_size is a size_t, so compare first in order to prevent an
     * underflow.
     */
    if(get_size(ptr) > req_size + HEADER_SIZE) {
        new_size = get_size(ptr) - req_size - HEADER_SIZE;
    } else {
        new_size = 0;
    }

#ifdef SPLITTING_FIXED
    min_split_size = MIN_SPLITTING_SIZE;
#endif /* SPLITTING_FIXED */
#ifdef SPLITTING_VARIABLE
    min_split_size = heap->dmm_knobs.min_split_size;
#endif /* SPLITTING_VARIABLE */

    if(new_size < min_split_size) {
        return;
    }

    new_block = (void *)((char *)ptr + req_size + HEADER_SIZE);
    new_block_size = get_size(ptr) - req_size - HEADER_SIZE;

    if(allocator->border_ptr == ptr) {
        allocator->border_ptr = new_block;
    }

    /* Resize the previous, to be used block */
    set_size_and_used(allocator, ptr, req_size);

#ifdef COALESCE_AFTER_SPLIT
    /* Try to coalesce with the next block if it is free */
#ifdef COALESCING_FIXED
    max_coal_size = MAX_COALESCE_SIZE;
#endif /* COALESCING_FIXED */
#ifdef COALESCING_VARIABLE
    max_coal_size = heap->dmm_knobs.max_coalesce_size;
#endif /* COALESCING_VARIABLE */

    if(next_block != NULL) {
        if(is_free(next_block) == true) {
#ifdef WITH_OWNERSHIP
            if(get_owner(next_block) == get_owner(ptr)) {
#endif /* WITH_OWNERSHIP */
                current_next_size = new_block_size + get_size(next_block) +
                    HEADER_SIZE;
                if(current_next_size <= max_coal_size) {
                    /* it's ok to coalesce the new split block with the next
                     * free block
                     */
                    remove_block_from_lists(&next_block, heap);
                    if(allocator->border_ptr == next_block) {
                        allocator->border_ptr = ptr;
                    }
                    new_block_size = current_next_size;
                }
#ifdef WITH_OWNERSHIP
            }
#endif /* WITH_OWNERSHIP */
        }
    }

#endif /* COALESCE_AFTER_SPLIT */

    set_size_and_free(allocator, new_block, new_block_size);

#ifdef WITH_FIXED_LISTS
    /* FIXME code from custom_free, some refactoring maybe?
     * Be careful, custom_free also does coalescing, we don't need that
     */

    /* Check if the block could be put in a fixed list */
    fixed_list_id = map_size_to_list(heap, new_block_size);

    if(fixed_list_id != -1) { /* put it in the right fixed list */
        current_maptable_node = heap->maptable_head;		
        if(fixed_list_id != 0) {
            for(i = 1; i < fixed_list_id; i++) {
                current_maptable_node = current_maptable_node->next;
            }
        }
#ifdef COUNT_HOPS
#ifdef FIFO_SORT_POLICY
        add_block(heap, &new_block, &current_maptable_node->fixed_list_head, &current_maptable_node->fixed_list_tail);
#else /* FIFO_SORT_POLICY */
        add_block(heap, &new_block, &current_maptable_node->fixed_list_head);
#endif /* FIFO_SORT_POLICY */
#else /* COUNT_HOPS */
#ifdef FIFO_SORT_POLICY
        add_block(&new_block, &current_maptable_node->fixed_list_head, &current_maptable_node->fixed_list_tail);
#else /* FIFO_SORT_POLICY */
        add_block(&new_block, &current_maptable_node->fixed_list_head);
#endif /* FIFO_SORT_POLICY */
#endif /* COUNT_HOPS */
    } else { /* put it in the free list */
#endif /* WITH_FIXED_LISTS */

#ifdef COUNT_HOPS
#ifdef FIFO_SORT_POLICY
        add_block(heap, &new_block, &heap->free_list_head, &heap->free_list_tail);
#else /* FIFO_SORT_POLICY */
        add_block(heap, &new_block, &heap->free_list_head);
#endif /* FIFO_SORT_POLICY */
#else /* COUNT_HOPS */
#ifdef FIFO_SORT_POLICY
        add_block(&new_block, &heap->free_list_head, &heap->free_list_tail);
#else /* FIFO_SORT_POLICY */
        add_block(&new_block, &heap->free_list_head);
#endif /* FIFO_SORT_POLICY */
#endif /* COUNT_HOPS */
#ifdef WITH_FIXED_LISTS
    }
#endif /* WITH_FIXED_LISTS */

#ifdef WITH_OWNERSHIP
    set_owner(new_block, heap);
#endif /* WITH_OWNERSHIP */

}
