/*
 *   Copyright 2011 Institute of Communication and Computer Systems (ICCS) 
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */

#include "sys_alloc.h"

#ifdef WITH_MMAP
#include <sys/stat.h> /* for open() */
#include <fcntl.h> /* for open() */

#include <sys/mman.h>
#else /* WITH_MMAP */
#include <unistd.h> /* for sbrk() */
#endif /* WITH_MMAP */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#ifdef HAVE_LOCKS
#include "posix_lock.h"
#endif /* HAVE_LOCKS */
#include "other.h"
#include "block_header.h"
#ifdef WITH_STATS
#include <dmmlib/print_stats.h>
#endif /* WITH_STATS */

#ifdef WITH_MMAP
static int dev_zero_fd = -1; /* Cached file descriptor for /dev/zero. */
#endif /* WITH_MMAP */

void *sys_alloc(allocator_t *allocator, heap_t *heap, size_t size) {

    size_t allocation_size, previous_size, previous_size_availability;
    void *ptr;
#ifdef WITH_MMAP
    int fd;
#endif /* WITH_MMAP */

#ifdef HAVE_LOCKS
    sbrk_lock();
#endif /* HAVE_LOCKS */

    allocation_size = req_padding(size) + HEADER_SIZE;

    /* Step 1: Check if the memory allocator is initialized.
     * 
     * Before the first application malloc() call, the allocator uses
     * sys_alloc() to allocate space for its metadata. So, these are the first
     * data blocks in the heaps and there are no previous data blocks.
     */
    if(allocator->initialized == false) {
        previous_size = 0;
        previous_size_availability = 1; /* Occupied and of 0 size */
        allocator->initialized = true;
    } else {
        previous_size = get_size(allocator->border_ptr);
        previous_size_availability = get_size_availability(allocator->border_ptr);
    }

    if(allocation_size > allocator->remaining_size) {
#ifndef NO_SYSTEM_CALLS
        if(allocation_size < SYS_ALLOC_SIZE) {
            allocation_size = SYS_ALLOC_SIZE;
        } else {
            /* If allocation size is more than the default size, then request
             * multiples of it.
             */
            allocation_size = ((size + HEADER_SIZE) / SYS_ALLOC_SIZE + 1) *
                SYS_ALLOC_SIZE;
        }
#ifdef WITH_MMAP
        if(dev_zero_fd < 0) {
            dev_zero_fd = open("/dev/zero", O_RDWR);
        }
        fd = dev_zero_fd;
        ptr = mmap(0, allocation_size, PROT_READ|PROT_WRITE,
                MAP_SHARED, fd, 0);
#endif /* WITH_MMAP */
#ifdef WITH_SBRK
        ptr = sbrk((int) allocation_size);
#endif /* WITH_SBRK */

        /* Print a warning for sbrk / mmap in case the user has enabled
         * coalescing support
         */
#if defined (COALESCING_FIXED) || defined (COALESCING_VARIABLE)
        if(allocator->border_ptr != NULL && ptr != (void *) 
                ((char *) allocator->border_ptr + previous_size)) {
            printf("sbrk() / mmap() does not return sequential space. You should"
                    " disable coalescing.\n");
        }
#endif /* COALESCING_FIXED || COALESCING_VARIABLE */    

        if(ptr == (void *)-1) {
#endif /* NO_SYSTEM_CALLS */
            printf("Couldn't allocate more space from the system.\n");
#ifdef WITH_STATS
            print_stats(allocator);
#endif /* WITH_STATS */
            return NULL;
#ifndef NO_SYSTEM_CALLS
        } else {
            allocator->remaining_size += allocation_size;
        }
#endif /* NO_SYSTEM_CALLS */
    } else {
        ptr = (void *)((char *)allocator->border_ptr + previous_size);
        allocator->remaining_size -= allocation_size;
    }

    /* Go to the data part of the block */
    ptr = (void *) ((char *) ptr + HEADER_SIZE);
    
    /* Set the new border pointer */
    allocator->border_ptr = ptr;

    /* Set some values for block header */
    set_size_and_used(allocator, ptr, req_padding(size));
    set_previous_size_availability(ptr, previous_size_availability);

#ifdef REQUEST_SIZE_INFO
    set_requested_size(ptr, size);
#endif /* REQUEST_SIZE_INFO */

#ifdef FUTURE_FEATURES
    /* Update the used blocks list */
    push_block(&ptr, &heap->used_blocks_head);
#endif /* FUTURE_FEATURES */

#ifdef WITH_STATS
    /* Iraklis' request: Each sys_alloc() call is equivalent to 2 hops. */
#ifdef COUNT_HOPS
    heap->dmm_stats.total_hops += 2;
#endif /* COUNT_HOPS */   
    /* Update statistics */
#ifdef REQUEST_SIZE_INFO
    heap->dmm_stats.mem_requested += size;
#endif /* REQUEST_SIZE_INFO */
    heap->dmm_stats.mem_used += HEADER_SIZE + req_padding(size);
    heap->dmm_stats.mem_allocated += HEADER_SIZE + req_padding(size);
    heap->dmm_stats.live_objects += 1;
    heap->dmm_stats.num_malloc += 1;
#endif /* WITH_STATS */

    /* FIXME To be refactored - END */

#ifdef HAVE_LOCKS
    sbrk_unlock();
#endif /* HAVE_LOCKS */

    return  ptr;
}

